/*
 * LEAF Visual Novel System For X
 * (c) Copyright 1999,2000 Go Watanabe mailto:go@denpa.org
 * All rights reserverd.
 *
 * ORIGINAL LVNS (c) Copyright 1996-1999 LEAF/AQUAPLUS Inc.
 *
 * $Id: LvnsCoreSound.c,v 1.29 2001/08/02 20:05:30 go Exp $
 *
 */

/*
 * LvnsCore ɽ
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/socket.h>

#include "LvnsCoreP.h"

/*
 * ɥץ̿󥸥
 * ɤαդλȡҥץ̿뤿ᤳƤФ
 */
static void
SoundControl(XtPointer cl, int *fd, XtInputId *id)
{
    char buf[100];
    size_t len;
    LvnsCoreWidget lcw = (LvnsCoreWidget) cl;

    if ((len = read(LCW.sound_ctl_fd, buf, 1)) < 0) {
        perror("SoundControl");
    } else if (len == 1) {
        if (buf[0] == 'E') {
            /* ä˽̵ */
    if (LCW.sound_ctl_id) {
        XtRemoveInput(LCW.sound_ctl_id);
    }
    LCW.sound_ctl_id = None;
        }
    }
}

/*
 * ǡɤ߹ 
 */
static void
LoadSound(LvnsCoreWidget lcw, const char* name)
{
    char buf[100];
    if (LCW.sound_ctl_fd >= 0 && LCW.sound_ctl_id == None) {
        snprintf(buf, sizeof buf, "L%s\n", name);
        write(LCW.sound_ctl_fd, buf, strlen(buf));
    }
}

/*
 * ճ
 */
static void
StartSound(LvnsCoreWidget lcw)
{
    /* ʣΥɤƱĤ餻ʤ */
    if (LCW.sound_ctl_fd >= 0 && LCW.sound_ctl_id == None) {
        XtAppContext app = XtWidgetToApplicationContext((Widget)lcw);
        write(LCW.sound_ctl_fd, "S\n", 2);
        LCW.sound_ctl_id = XtAppAddInput(app, 
                                         LCW.sound_ctl_fd,
                                         (XtPointer)XtInputReadMask,
                                         SoundControl,
                                         lcw);
    }
}

/*
 * մλ(ǡ˴)
 */
static void
StopSound(LvnsCoreWidget lcw)
{
    dprintf((stderr, "stop sound\n"));
    if (LCW.sound_ctl_fd >= 0 && LCW.sound_ctl_id) {
        write(LCW.sound_ctl_fd, "E\n", 2);
    }
}

/*
 * վ֤γǧ
 */
static int
GetSoundState(LvnsCoreWidget lcw)
{
    if (LCW.sound_ctl_fd >= 0 && LCW.sound_ctl_id) {
        return True;
    } else {
        return False;
    }
}       

/* 
 * ե󥸥
 */
static void
CloseSound(LvnsCoreWidget lcw)
{
    /* ѥץ */
    if (LCW.sound_ctl_id != None) {
        XtRemoveInput(LCW.sound_ctl_id);
    }
    if (LCW.sound_ctl_fd >= 0) {
        close(LCW.sound_ctl_fd);  /* Ϥ򤦤ƽλ */
    }

    LCW.sound_ctl_id  = None;
    LCW.sound_ctl_fd  = -1;
}

/*
 * ɽѻҥץ 
 */
#include "audioinfo.h"  /* SOUND-audio Control */
static void
OpenSound(LvnsCoreWidget lcw)
{
    Lvns *lvns = LCW.info;
    char *devicename;
    AUDIOInfo *ainfo;
    int fd[2];
    pid_t pid;

    devicename = getenv("AUDIODEVICE");
    if (!devicename)
        devicename = AUDIODEVICE;

    if ((ainfo = audioinfo_new(devicename)) == NULL) {
        return;
    }

    if (socketpair(AF_UNIX, SOCK_STREAM, AF_UNSPEC, fd) < 0) {
        perror("socketpair AF_UNIX/PF_UNSPEC");
        return;
    }

    if ((pid = fork()) < 0) {
        perror("fork");
        audioinfo_delete(ainfo);
        close(fd[0]);
        close(fd[1]);
        return;
    } else if (pid) {
        audioinfo_delete(ainfo);
        close(fd[1]);
        LCW.sound_ctl_fd = fd[0];
        return;
    }

    /* սѻҥץ */
    dprintf((stderr, "starting sound process...\n"));
    close(fd[0]);

    {
        char buf[100];
        u_char *sound_data = NULL;
        size_t sound_len;
        FILE *fp;

        if ((fp = fdopen(fd[1], "rw")) == NULL) {
            perror("fdopen");
            exit(1);
        }
        
        while (fgets(buf, sizeof buf, fp)) {
            
            char *p;
            if ((p = strchr(buf, '\n'))) {
                *p = '\0';
            }

            switch (buf[0]) {
            case 'L':
                /*  */
                sound_data = LvnsLoadData(lvns, buf + 1, &sound_len);
                break;

            case 'S': 
                /* ճ */
                if (sound_data) {
                    audioinfo_play(ainfo, sound_data, sound_len);
                }
                /* մλƤ */
                write(fd[1], "E", 1);
                break;

            case 'E':
                /* սλ */
                if (sound_data)
                    free(sound_data);
                sound_data = NULL;
            }
        }
    }
    exit(0);
}

void 
LvnsCoreInitSound(LvnsCoreWidget lcw)
{
	LCW.sound.depend   = lcw;
	LCW.sound.open     = OpenSound;
	LCW.sound.close    = CloseSound;
	LCW.sound.load     = LoadSound;
	LCW.sound.start    = StartSound;
	LCW.sound.stop     = StopSound;
	LCW.sound.getState = GetSoundState;

    LCW.sound_ctl_id  = None;
    LCW.sound_ctl_fd  = -1;
}
